import os
import sys
import shlex
import asyncio
import webbrowser
import subprocess
import importlib.resources as resources

from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse, HTMLResponse, FileResponse, StreamingResponse
from fastapi.staticfiles import StaticFiles
import uvicorn
from pydantic import BaseModel

import buelon
from buelon.hub import BiWorkerClient, settings, compressed_message_to_steps

app = FastAPI()
worker_client: BiWorkerClient | None = None
settings.worker.info['name'] = f"Web App ({settings.worker.info['name']})"


def get_static_file(filename: str) -> str:
    # Opens the file as text
    with resources.files("buelon.static").joinpath(filename).open("r", encoding="utf-8") as f:
        return f.read()


def get_static_path(filename: str) -> str:
    # Gets the actual filesystem path (works even if installed in venv)
    return str(resources.files("buelon.static").joinpath(filename))


@app.get("/", response_class=HTMLResponse)
async def index():
    return get_static_file("index.html")


@app.get("/static/{path:path}")
async def static_file(path: str):
    return FileResponse(get_static_path(path))


@app.post("/data")
async def get_data():
    global worker_client
    assert worker_client is not None
    try:
        data = await worker_client.get_web_info(True)
    except:
        worker_client = await worker_client.__aenter__()
        data = await worker_client.get_web_info(True)
    return JSONResponse(content=data)


@app.post("/errors")
async def get_errors():
    global worker_client
    assert worker_client is not None
    try:
        compressed_jobs, error_info = await worker_client.errors()
    except:
        worker_client = await worker_client.__aenter__()
        compressed_jobs, error_info = await worker_client.errors()
    jobs = compressed_message_to_steps(compressed_jobs)
    json_jobs = [job.to_json() for job in jobs]
    data = {
        "jobs": json_jobs,
        "errors": error_info
    }
    return JSONResponse(content=data)


@app.post('/reset-errors')
async def reset_errors():
    global worker_client
    assert worker_client is not None
    try:
        await worker_client.reset_errors()
    except:
        worker_client = await worker_client.__aenter__()
        await worker_client.reset_errors()
    return JSONResponse(content={'status': 'success'})


class Job(BaseModel):
    id: str


@app.post("/job-parents-and-results")
async def api_job_parents_and_results(job: Job):
    global worker_client
    assert worker_client is not None
    try:
        data = await worker_client.get_job_parents_and_results(job.id)
    except:
        worker_client = await worker_client.__aenter__()
        data = await worker_client.get_job_parents_and_results(job.id)
    return JSONResponse(content=data)


async def stream_subprocess_logs(cmd_parts):
    """
    Asynchronously runs a command and streams its stdout and stderr.
    This version uses an asyncio.Queue to correctly merge the two streams.
    """
    process = await asyncio.create_subprocess_exec(
        *cmd_parts,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )

    queue = asyncio.Queue()

    # This task reads from a stream (stdout or stderr) and puts lines into the queue.
    async def reader_task(stream):
        while True:
            line = await stream.readline()
            if not line:
                break
            await queue.put(line)
        # When the stream is finished, put a sentinel value in the queue.
        await queue.put(None)

    # Start tasks for both stdout and stderr
    stdout_reader = asyncio.create_task(reader_task(process.stdout))
    stderr_reader = asyncio.create_task(reader_task(process.stderr))

    finished_streams = 0
    while finished_streams < 2:
        line = await queue.get()
        if line is None:
            # A stream has finished.
            finished_streams += 1
        else:
            yield line.decode('utf-8')

    # Wait for the process and reader tasks to complete.
    await process.wait()
    await asyncio.gather(stdout_reader, stderr_reader)


@app.post("/run-job")
async def api_run_job(job: Job):
    # # assert worker_client is not None
    # # Construct the command to run the buelon worker for a specific job
    # # Using repr() ensures the job.id string is correctly quoted
    # # The `-u` flag is important for unbuffered output, which is ideal for streaming
    # import_txt = 'import buelon'
    # new_name = f"Web App Subprocess ({settings.worker.info['name']})"
    # rename_worker_txt = f"buelon.settings.settings.worker.info['name'] = {repr(new_name)}"
    # execution_txt = f'buelon.worker.work({repr(job.id)})'
    # command = f'{sys.executable} -u -c "{import_txt};{rename_worker_txt};{execution_txt}"'
    import_txt = 'import buelon'
    new_name = f"Web App Subprocess ({settings.worker.info['name']})"
    rename_worker_txt = f"buelon.settings.settings.worker.info['name'] = {repr(new_name)}"
    execution_txt = f'buelon.worker.work({repr(job.id)})'

    # Build the Python code first, then properly quote it for the shell
    python_code = f"{import_txt};{rename_worker_txt};{execution_txt}"
    command = f'{sys.executable} -u -c {shlex.quote(python_code)}'

    cmd_parts = shlex.split(command)

    return StreamingResponse(stream_subprocess_logs(cmd_parts), media_type="text/plain")


async def start_app(open_browser: bool = False):
    global worker_client
    try:
        port = int(os.environ.get('BOO_WEB_PORT', 11011))
    except:
        port = 11011

    host = os.environ.get('BOO_WEB_HOST', "localhost")

    if open_browser:
        webbrowser.open(f"http://{host}:{port}")

    # async with BiWorkerClient() as client:
    #     worker_client = client
    #     # Start FastAPI with uvicorn
    #     config = uvicorn.Config(app, host=host, port=port, log_level="info")
    #     server = uvicorn.Server(config)
    #     await server.serve()

    worker_client = BiWorkerClient()
    worker_client = await worker_client.__aenter__()

    try:
        # Start FastAPI with uvicorn
        config = uvicorn.Config(app, host=host, port=port, log_level="info")
        server = uvicorn.Server(config)
        await server.serve()
    except:
        try:
            await worker_client.__aexit__(*sys.exc_info())
        except: pass


def run(open_browser: bool = False):
    asyncio.run(start_app(open_browser=open_browser))


if __name__ == "__main__":
    asyncio.run(start_app(open_browser=True))





# import sys
# import asyncio
# import webbrowser
# import importlib.resources as resources
#
# from unsync import unsync
# from flask import Flask, request, jsonify, send_file, render_template_string
# from buelon.hub import BiWorkerClient
#
# app = Flask(__name__)
# worker_client = BiWorkerClient()
#
#
# def get_static_file(filename: str) -> str:
#     # Opens the file as text
#     with resources.files("buelon.static").joinpath(filename).open("r", encoding="utf-8") as f:
#         return f.read()
#
#
# def get_static_path(filename: str) -> str:
#     # Gets the actual filesystem path (works even if installed in venv)
#     return str(resources.files("buelon.static").joinpath(filename))
#
#
# @app.route("/")
# def index():
#     return render_template_string(get_static_file("index.html"))
#
#
# @app.route("/static/<path:path>")
# def static_file(path):
#     return send_file(get_static_path(path))
#
#
# @app.route('/data', methods=['POST'])
# def get_data():
#     @unsync
#     async def get_data_sync():
#         return await worker_client.get_web_info(True)
#
#     return jsonify(get_data_sync().result())
#
#
# def run(open_browser: bool = False):
#     _run(open_browser).result()
#
#
# @unsync
# async def _run(open_browser: bool = False):
#     global worker_client
#     port = 11011
#     host = 'localhost'
#
#     if open_browser:  # ('-y' in sys.argv and '-n' not in sys.argv) or f'{input("Open Browser? (y/n)")}'.lower().startswith('y'):
#         webbrowser.open(f'http://{host}:{port}')
#
#     async with BiWorkerClient() as client:
#         worker_client = client
#         app.run(port=port, host=host)
#
#
# if __name__ == '__main__':
#     run()
#
#
#
#



import os
import inspect
import uuid
import sqlite3
import time

import buelon.helpers.sqlite3_helper
import buelon.helpers.postgres

DEBUG_TABLE = 'debug'

# pg = buelon.helpers.postgres.get_postgres_from_env()


def counter(_id: str, change: int | float = 1, table_name: str = 'counter'):
    """
    a counter that can be used to track the progress of a task

    Args:
        _id: the id of the counter
        change (default: 1): the amount to change the counter by
        table_name (default: 'counter'): the name of the table to use

    Returns:
        the current count of the counter
    """
    db = buelon.helpers.sqlite3_helper.Sqlite3('test.db')
    try:
        db.query(f'update {table_name} set count = count + {change} where id = ?;', _id)
        t = db.download_table(sql=f'select * from {table_name} where id = ?;', parameters=(_id,))
        # pg.query(f'update {table_name} set count = count + {change} where id = \'{_id}\';')
        # t = pg.download_table(sql=f'select * from {table_name} where id = \'{_id}\';')
    except:
        t = []
    if not t:
        t.append({'id': _id, 'count': change})
        db.upload_table(table_name, t, id_column='id')
        # pg.upload_table(table_name, t, id_column='id')
        return change
    row = t[0]
    # row['count'] += change
    # db.upload_table(table_name, t, id_column='id')
    return row['count']


def timeit(func: callable):
    return func
    module_name, func_name = os.path.basename(inspect.getmodule(func).__file__).split('.')[0], func.__name__

    def inner(*args, **kwargs):
        t = time.time()
        v = func(*args, **kwargs)
        counter(f'{module_name}.{func_name}', time.time() - t, DEBUG_TABLE)
        return v

    async def inner_async(*args, **kwargs):
        t = time.time()
        if inspect.isasyncgenfunction(func):
            async def gen():
                nonlocal t
                async for v in func(*args, **kwargs):
                    yield v
                counter(f'{module_name}.{func_name}', time.time() - t, DEBUG_TABLE)

            return gen()
        v = await func(*args, **kwargs)
        counter(f'{module_name}.{func_name}', time.time() - t, DEBUG_TABLE)
        return v

    return inner if not inspect.iscoroutinefunction(func) else inner_async


def time_from_here():
    t = time.time()
    db = buelon.helpers.sqlite3_helper.Sqlite3('test.db')
    i = f'{uuid.uuid4()}'
    db.upload_table('tags', [{'id': i, 'time': t}], id_column='id')
    return i


def time_to_here(_id: str, tag: str):
    db = buelon.helpers.sqlite3_helper.Sqlite3('test.db')
    t = []
    try:
        t = db.download_table('delete from tags where id = ? returning *', _id)
    except sqlite3.OperationalError:
        pass
    if t:
        counter(tag, time.time() - t[0]['time'], DEBUG_TABLE)



import sys
import shlex
import asyncio
import webbrowser
import subprocess
import importlib.resources as resources

from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse, HTMLResponse, FileResponse
from fastapi.staticfiles import StaticFiles
import uvicorn
from pydantic import BaseModel

import buelon
from buelon.hub import WorkerClient, settings, compressed_message_to_steps

app = FastAPI()
worker_client: WorkerClient | None = None
settings.worker.info['name'] = f"Web App ({settings.worker.info['name']})"


def get_static_file(filename: str) -> str:
    # Opens the file as text
    with resources.files("buelon.static").joinpath(filename).open("r", encoding="utf-8") as f:
        return f.read()


def get_static_path(filename: str) -> str:
    # Gets the actual filesystem path (works even if installed in venv)
    return str(resources.files("buelon.static").joinpath(filename))


@app.get("/", response_class=HTMLResponse)
async def index():
    return get_static_file("index.html")


@app.get("/static/{path:path}")
async def static_file(path: str):
    return FileResponse(get_static_path(path))


@app.post("/data")
async def get_data():
    assert worker_client is not None
    data = await worker_client.get_web_info(True)
    return JSONResponse(content=data)


@app.post("/errors")
async def get_errors():
    assert worker_client is not None
    compressed_jobs, error_info = await worker_client.errors()
    jobs = compressed_message_to_steps(compressed_jobs)
    json_jobs = [job.to_json() for job in jobs]
    data = {
        "jobs": json_jobs,
        "errors": error_info
    }
    return JSONResponse(content=data)


class Job(BaseModel):
    id: str


@app.post("/job-parents-and-results")
async def api_job_parents_and_results(job: Job):
    assert worker_client is not None
    data = await worker_client.get_job_parents_and_results(job.id)
    return JSONResponse(content=data)


@app.post("/run-job")
async def api_run_job(job: Job):
    assert worker_client is not None
    cmd = f'{sys.executable} -c "import buelon;buelon.worker.work({repr(job.id)})"'
    cmd = shlex.split(cmd)
    p = subprocess.Popen(cmd)


async def start_app(open_browser: bool = False):
    global worker_client
    port = 11011
    host = "localhost"

    if open_browser:
        webbrowser.open(f"http://{host}:{port}")

    # async with WorkerClient() as client:
    #     worker_client = client
    #     # Start FastAPI with uvicorn
    #     config = uvicorn.Config(app, host=host, port=port, log_level="info")
    #     server = uvicorn.Server(config)
    #     await server.serve()

    worker_client = WorkerClient()
    worker_client = await worker_client.__aenter__()

    try:
        # Start FastAPI with uvicorn
        config = uvicorn.Config(app, host=host, port=port, log_level="info")
        server = uvicorn.Server(config)
        await server.serve()
    except:
        try:
            await worker_client.__aexit__(*sys.exc_info())
        except: pass


def run(open_browser: bool = False):
    asyncio.run(start_app(open_browser=open_browser))


if __name__ == "__main__":
    asyncio.run(start_app(open_browser=True))





# import sys
# import asyncio
# import webbrowser
# import importlib.resources as resources
#
# from unsync import unsync
# from flask import Flask, request, jsonify, send_file, render_template_string
# from buelon.hub import WorkerClient
#
# app = Flask(__name__)
# worker_client = WorkerClient()
#
#
# def get_static_file(filename: str) -> str:
#     # Opens the file as text
#     with resources.files("buelon.static").joinpath(filename).open("r", encoding="utf-8") as f:
#         return f.read()
#
#
# def get_static_path(filename: str) -> str:
#     # Gets the actual filesystem path (works even if installed in venv)
#     return str(resources.files("buelon.static").joinpath(filename))
#
#
# @app.route("/")
# def index():
#     return render_template_string(get_static_file("index.html"))
#
#
# @app.route("/static/<path:path>")
# def static_file(path):
#     return send_file(get_static_path(path))
#
#
# @app.route('/data', methods=['POST'])
# def get_data():
#     @unsync
#     async def get_data_sync():
#         return await worker_client.get_web_info(True)
#
#     return jsonify(get_data_sync().result())
#
#
# def run(open_browser: bool = False):
#     _run(open_browser).result()
#
#
# @unsync
# async def _run(open_browser: bool = False):
#     global worker_client
#     port = 11011
#     host = 'localhost'
#
#     if open_browser:  # ('-y' in sys.argv and '-n' not in sys.argv) or f'{input("Open Browser? (y/n)")}'.lower().startswith('y'):
#         webbrowser.open(f'http://{host}:{port}')
#
#     async with WorkerClient() as client:
#         worker_client = client
#         app.run(port=port, host=host)
#
#
# if __name__ == '__main__':
#     run()
#
#
#
#



# Import modules from subpackages
from . import settings
from . import hub_v1, worker_v1, bucket_v1
from .core import action, execution, loop, pipe, pipe_interpreter, step, step_definition
from . import bucket, hub, worker, command_line
try:
    # from .cython import c_bucket, c_hub, c_worker
    from . import bucket as c_bucket, hub as c_hub, worker as c_worker
except ImportError:
    from . import bucket as c_bucket, hub as c_hub, worker as c_worker
from .helpers import json_parser, pipe_util, postgres, sqlite3_helper
from .examples import demo, example


# Define the public API
__all__ = ['bucket', 'hub', 'worker', 'command_line', 'core', 'cython', 'helpers', 'cython', 'examples']


"""Module for handling action checks in a pipeline system."""

from buelon.helpers import pipe_util
from . import pipe
from . import pipe_interpreter


def check_for_actions(index: int, line: str, variables: dict) -> bool:
    """Check and process action calls in a line of pipeline code.

    This function identifies and processes function calls in the pipeline code.
    It handles variable assignment from function returns and validates the syntax
    of the function call.

    Args:
        index (int): The current line index in the pipeline code.
        line (str): The current line of code being processed.
        variables (dict): A dictionary containing pipeline variables and functions.

    Returns:
        bool: True if an action was found and processed, False otherwise.

    Raises:
        SyntaxError: If there are syntax errors in the function call.
        TypeError: If the called object is not a Pipe instance.

    The function performs the following steps:
    1. Checks if the line contains a function call.
    2. Validates the syntax of the function call.
    3. Extracts the return value assignment if present.
    4. Parses the function name and arguments.
    5. Validates the function exists in the variables.
    6. Executes the function and assigns the return value if necessary.
    """
    begin_token, end_token = pipe_interpreter.PIPE_TOKENS['func']
    if begin_token in line and end_token in line and not line.strip().startswith('for'):
        # Validation and parsing logic...
        if line.count(begin_token) > 1:
            raise SyntaxError(f'Line {index + 1}: Invalid call. To many \'{begin_token}\'')
        if line.count(end_token) > 1:
            raise SyntaxError(f'Line {index + 1}: Invalid call. To many \'{end_token}\'')
        return_value = ''
        if '=' in line.replace('=>', ''):
            if line.count('=') > 1:
                raise SyntaxError(f'Line {index + 1}: Invalid call. To many \'=\'')
            return_value, line = line.split('=')
            return_value = pipe_util.trim(return_value)
        name, args = line.split(begin_token)
        args = ','.join(
            [pipe_util.trim(arg) for arg in args.split(end_token)[0].strip().split(',') if pipe_util.trim(arg) != ''])
        name = pipe_util.trim(name)
        if name not in variables:
            raise SyntaxError(f'Line {index + 1}: Call to unknown function \'{name}\'')

        # Execute the function
        _pipe: pipe.Pipe = variables[name]
        if not isinstance(_pipe, pipe.Pipe):
            print('pipe is ',  type(_pipe))
            raise TypeError(f'Line {index+1}: \'{name}\' is not a function')
        val = _pipe.create_steps(index, variables, args, pipe_util.json_copy(_pipe.kwargs))

        # Assign return value if necessary
        if return_value:
            variables[return_value] = val

        return True
    return False
